// Copyright 2002 Adam Megacz, see the COPYING file for licensing [LGPL]
#include "GCJ.cc"

// we have to do this because the jpeg libraries use the symbol 'INT32'
#define INT32 WIN32_INT32

// this has to precede the others so we don't get collisions on min/max
#include <org/xwt/Box.h>

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>

#include <windows.h>
#include <mmsystem.h>
#undef STRICT
#undef MAX_PRIORITY
#undef MIN_PRIORITY
#undef NORM_PRIORITY

#include <gcj/cni.h>

#include <java/lang/Integer.h>
#include <java/util/Hashtable.h>
#include <org/xwt/Box.h>
#include <org/xwt/Surface.h>
#include <org/xwt/DoubleBuffer.h>
#include <org/xwt/Picture.h>
#include <org/xwt/ByteStream.h>
#include <org/xwt/Platform.h>
#include <org/xwt/Platform$ParsedFont.h>
#include <org/xwt/plat/Win32.h>
#include <org/xwt/plat/Win32$Win32Font.h>
#include <org/xwt/plat/Win32$Win32Surface.h>
#include <org/xwt/plat/Win32$Win32DoubleBuffer.h>
#include <org/xwt/plat/Win32$Win32Picture.h>
#include <org/xwt/util/Log.h>
#include <org/xwt/util/Semaphore.h>

// for debugging
#include <java/lang/System.h>
#include <java/io/PrintStream.h>

#define WM_USER_SETCURSOR WM_USER
#define WM_USER_DISPOSE (WM_USER + 1)
#define WM_USER_CREATEWINDOW (WM_USER + 2)
#define WS_NORMAL (WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)

// FEATURE: there are lots of places where HANDLE's get casted to jint's -- this will break on Win64
//          a clean way to do this would be to '#define jraw (gnu::gcj::RawData*)'

// Callbacks ////////////////////////////////////////////////////////////////////

LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
    org::xwt::plat::Win32$Win32Surface* surface =
        (org::xwt::plat::Win32$Win32Surface*)org::xwt::plat::Win32::hwndToWin32SurfaceMap->get(new java::lang::Integer((jint)hwnd));

    if (surface != NULL) {
        return (LRESULT)surface->WndProc((jint)hwnd, (jint)iMsg, (jint)wParam, (jint)lParam);

    } else {
        // this is really lame -- Win32 insists on being able to call your WndProc BEFORE CreateWindow returns...
        return DefWindowProc(hwnd, iMsg, wParam, lParam);
    }
}

// This function iterates over each family (lparam == 0), and then over each size (lparam == 1)
int CALLBACK fontproc(const LOGFONTA* enumlogfont, const TEXTMETRICA* tm, long unsigned int type, LPARAM lparam) {

    if (lparam == 0) {
        LOGFONT lf;
        lf.lfCharSet = ANSI_CHARSET;
        strncpy(lf.lfFaceName, enumlogfont->lfFaceName, 32);
        lf.lfPitchAndFamily = 0;
        EnumFontFamiliesEx((HDC)org::xwt::plat::Win32::desktop_dc, &lf, fontproc, 1, 0);

    } else {
        org::xwt::plat::Win32::addFont(JvNewStringLatin1(enumlogfont->lfFaceName),
                                       ((type & RASTER_FONTTYPE) == 0) ? 0 : tm->tmHeight,
                                       tm->tmItalic == 0 ? 0 : 1, 
                                       tm->tmWeight <= 400 ? 0 : 1);
    }
    return 1;
}


// Initialization ////////////////////////////////////////////////////////////////////

static int window_class_counter = 0;

jstring org::xwt::plat::Win32::getTempPath() {
    char buf[1024];
    DWORD ret = GetTempPath(1024, buf);
    if (ret == 0) criticalAbort(JvNewStringLatin1("GetTempPath() failed"));
    return JvNewStringLatin1(buf);
}

// XOR mask for the hand cursor
static unsigned char hand_cursor_xor[32 * 4] = {
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x0C, 0x00, 0x00,
  0x00, 0x0C, 0x00, 0x00,
  0x00, 0x0C, 0x00, 0x00,
  0x00, 0x0C, 0x00, 0x00,
  0x00, 0x0C, 0x00, 0x00,
  0x00, 0x0D, 0x80, 0x00,
  0x00, 0x0D, 0xB0, 0x00,
  0x00, 0x0D, 0xB4, 0x00,
  0x00, 0x0D, 0xB6, 0x00,
  0x00, 0xCF, 0xF6, 0x00,
  0x00, 0xEF, 0xFE, 0x00,
  0x00, 0x6F, 0xFE, 0x00,
  0x00, 0x2F, 0xFE, 0x00,
  0x00, 0x3F, 0xFE, 0x00,
  0x00, 0x1F, 0xFE, 0x00,
  0x00, 0x1F, 0xFC, 0x00,
  0x00, 0x0F, 0xFC, 0x00,
  0x00, 0x0F, 0xFC, 0x00,
  0x00, 0x07, 0xF8, 0x00,
  0x00, 0x07, 0xF8, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00
};

// AND mask for the hand cursor
static unsigned char hand_cursor_and[32 * 4] = {
  0xFF, 0xF3, 0xFF, 0xFF,
  0xFF, 0xE1, 0xFF, 0xFF,
  0xFF, 0xE1, 0xFF, 0xFF,
  0xFF, 0xE1, 0xFF, 0xFF,
  0xFF, 0xE1, 0xFF, 0xFF,
  0xFF, 0xE0, 0x7F, 0xFF,
  0xFF, 0xE0, 0x0F, 0xFF,
  0xFF, 0xE0, 0x03, 0xFF,
  0xFF, 0xE0, 0x01, 0xFF,
  0xFF, 0x20, 0x00, 0xFF,
  0xFE, 0x00, 0x00, 0xFF,
  0xFE, 0x00, 0x00, 0xFF,
  0xFF, 0x00, 0x00, 0xFF,
  0xFF, 0x80, 0x00, 0xFF,
  0xFF, 0x80, 0x00, 0xFF,
  0xFF, 0xC0, 0x00, 0xFF,
  0xFF, 0xC0, 0x01, 0xFF,
  0xFF, 0xE0, 0x01, 0xFF,
  0xFF, 0xE0, 0x01, 0xFF,
  0xFF, 0xF0, 0x03, 0xFF,
  0xFF, 0xF0, 0x03, 0xFF,
  0xFF, 0xF0, 0x03, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0xFF
};


void org::xwt::plat::Win32::natInit() {

    // grab desktop dc/handle
    desktop_handle = (jint)GetDesktopWindow();
    desktop_dc = (jint)GetDC((HWND)desktop_handle);

    // create cursors
    org::xwt::plat::Win32::wait_cursor = (jint)LoadCursor(NULL, IDC_WAIT);
    org::xwt::plat::Win32::default_cursor = (jint)LoadCursor(NULL, IDC_ARROW);
    org::xwt::plat::Win32::crosshair_cursor = (jint)LoadCursor(NULL, IDC_CROSS);
    org::xwt::plat::Win32::text_cursor = (jint)LoadCursor(NULL, IDC_IBEAM);
    org::xwt::plat::Win32::move_cursor = (jint)LoadCursor(NULL, IDC_SIZEALL);
    org::xwt::plat::Win32::sizenesw_cursor = (jint)LoadCursor(NULL, IDC_SIZENESW);
    org::xwt::plat::Win32::sizens_cursor = (jint)LoadCursor(NULL, IDC_SIZENS);
    org::xwt::plat::Win32::sizenwse_cursor = (jint)LoadCursor(NULL, IDC_SIZENWSE);
    org::xwt::plat::Win32::sizewe_cursor = (jint)LoadCursor(NULL, IDC_SIZEWE);
    org::xwt::plat::Win32::hand_cursor = (jint)CreateCursor(GetModuleHandle(NULL), 14, 1, 32, 32, hand_cursor_and, hand_cursor_xor);

    // enumerate fonts
    LOGFONT lf;
    lf.lfCharSet = ANSI_CHARSET;
    lf.lfFaceName[0] = 0;
    lf.lfPitchAndFamily = 0;
    EnumFontFamiliesEx((HDC)desktop_dc, &lf, fontproc, 0, 0);

    messagePumpThread = (jint)GetCurrentThreadId();
    messagePumpStarted->release();

    MSG msg;
    while(GetMessage(&msg, (HWND)NULL, 0, 0) > 0) {

        if (msg.message == WM_USER_CREATEWINDOW) {
            org::xwt::plat::Win32$Win32Surface *surface = (org::xwt::plat::Win32$Win32Surface*)msg.lParam;

            // we must create a unique window class name for each
            // window so that minimization icons can be set independantly
            char buf[255];
            sprintf(buf, "XWT_WINDOW_CLASS_%i", window_class_counter++);

            WNDCLASSEX wc;
            wc.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
            wc.lpfnWndProc = WndProc;
            wc.cbClsExtra = 0;
            wc.cbSize = sizeof(WNDCLASSEX);
            wc.cbWndExtra = 0;
            wc.hInstance = GetModuleHandle(NULL);
            wc.hIcon = NULL;
            wc.hIconSm = NULL;
            wc.hCursor = NULL;
            wc.hbrBackground = (HBRUSH)(COLOR_SCROLLBAR + 1);
            wc.lpszMenuName = NULL;
            wc.lpszClassName = buf;
            RegisterClassEx(&wc);
            
            surface->hwnd = (jint)CreateWindow(wc.lpszClassName, TEXT(""), msg.wParam ? WS_NORMAL : WS_POPUP, 200, 200, 100, 100,
                                               (HWND__*)NULL, (HMENU__*)NULL, GetModuleHandle(NULL), (LPVOID)NULL);
            SetFocus((HWND)surface->hwnd);
            surface->hwndCreated->release();
            
        } else {
            TranslateMessage(&msg);
            if (msg.message == WM_LBUTTONDOWN || msg.message == WM_RBUTTONDOWN || msg.message == WM_MBUTTONDOWN) SetFocus(msg.hwnd);
            DispatchMessage(&msg);
        }

    }
    java::lang::System::exit(-1);
}


// Platform Methods ///////////////////////////////////////////////////////////////////

jstring org::xwt::plat::Win32::_getEnv(jstring key) {
    int len = JvGetStringUTFLength(key);
    char buf[len + 1];
    JvGetStringUTFRegion(key, 0, len, buf);
    buf[len] = '\0';
    char buf2[1024];
    DWORD ret = GetEnvironmentVariable(buf, buf2, 1024);
    if (ret > 0 && ret < 1024) return JvNewStringLatin1(buf2);
    return NULL;
}

jstring org::xwt::plat::Win32::_fileDialog(jstring suggestedFileName, jboolean write) {

    char buf[1024];
    OPENFILENAME ofn;
    memset(buf, 0, 1024);
    memset(&ofn, 0, sizeof(OPENFILENAME));

    if (suggestedFileName != NULL)
        JvGetStringUTFRegion(suggestedFileName, 0, min(1023, JvGetStringUTFLength(suggestedFileName)), buf);
    
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.nMaxCustFilter = 0;
    ofn.lpstrFile = buf;
    ofn.nMaxFile = 1024;

    if (write) ofn.Flags |= OFN_OVERWRITEPROMPT;
    ofn.Flags |= OFN_HIDEREADONLY;

    int ret = write ? GetSaveFileName(&ofn) : GetOpenFileName(&ofn);
    return ret == 0 ? NULL : JvNewStringLatin1(buf);
}

void org::xwt::plat::Win32::__detectProxy(JArray<jstring>* container) {

    HKEY hkey;
    char buf[1024];
    DWORD buflen = 1024;
    DWORD type;
    LONG result = RegOpenKey(HKEY_CURRENT_USER, "Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings", &hkey);
    if (result != ERROR_SUCCESS) return;
    
    buf[0] = '\0';
    type = REG_SZ;
    buflen = 1024;
    result = RegQueryValueEx(hkey, "AutoConfigURL", NULL, &type, (LPBYTE)buf, &buflen);
    buf[buflen] = '\0';
    if (result == ERROR_SUCCESS) elements(container)[2] = JvNewStringLatin1(buf);

    buf[0] = '\0';
    type = REG_BINARY;
    RegQueryValueEx(hkey, "ProxyEnable", NULL, &type, (LPBYTE)buf, &buflen);
    if (buf[0] != 1) return;

    buf[0] = '\0';
    type = REG_SZ;
    buflen = 1024;
    RegQueryValueEx(hkey, "ProxyServer", NULL, &type, (LPBYTE)buf, &buflen);
    buf[buflen] = '\0';
    elements(container)[0] = JvNewStringLatin1(buf);

    buf[0] = '\0';
    buflen = 1024;
    RegQueryValueEx(hkey, "ProxyOverride", NULL, &type, (LPBYTE)buf, &buflen);
    buf[buflen] = '\0';
    elements(container)[1] = JvNewStringLatin1(buf);
}

jstring org::xwt::plat::Win32::_getClipBoard() {
    OpenClipboard((HWND)desktop_handle);
    HGLOBAL hmem = GetClipboardData(CF_TEXT);
    if (hmem == NULL) return NULL;
    char* buf = (char*)GlobalLock(hmem);
    if (buf == NULL) return NULL;
    jstring ret = JvNewStringLatin1(buf);
    GlobalUnlock(hmem);
    CloseClipboard();
    return ret;
}

void org::xwt::plat::Win32::_setClipBoard(jstring s) {
    OpenClipboard((HWND)desktop_handle);
    HGLOBAL hmem = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, JvGetStringUTFLength(s) + 1);
    if (hmem == NULL) return;
    char* buf = (char*)GlobalLock(hmem);
    if (buf == NULL) return;
    JvGetStringUTFRegion(s, 0, JvGetStringUTFLength(s), buf);
    buf[JvGetStringUTFLength(s)] = '\0';
    GlobalUnlock(hmem);
    SetClipboardData(CF_TEXT, hmem);
    CloseClipboard();
}

void org::xwt::plat::Win32::_criticalAbort(jstring message) {
    char buf[JvGetStringUTFLength(message) + 1];
    buf[JvGetStringUTFLength(message)] = '\0';
    JvGetStringUTFRegion(message, 0, JvGetStringUTFLength(message), buf);
    MessageBox (NULL, buf, "XWT Cannot Continue", MB_OK | MB_ICONSTOP | MB_TASKMODAL | MB_SETFOREGROUND);
    java::lang::System::exit(-1);
}

jint org::xwt::plat::Win32::_getScreenWidth() {
    RECT rect;
    SystemParametersInfo(SPI_GETWORKAREA, 0, &rect, 0);
    return rect.right - rect.left;
}

jint org::xwt::plat::Win32::_getScreenHeight() {
    RECT rect;
    SystemParametersInfo(SPI_GETWORKAREA, 0, &rect, 0);
    return rect.bottom - rect.top;
}

org::xwt::plat::Win32$Win32Font* org::xwt::plat::Win32::mapFont(org::xwt::Platform$ParsedFont* pf) {
    org::xwt::plat::Win32$Win32Font* ret = new org::xwt::plat::Win32$Win32Font();
    LOGFONT logfont;
    memset(&logfont, 0, sizeof(LOGFONT));
    logfont.lfHeight = -MulDiv(pf->size, GetDeviceCaps((HDC)org::xwt::plat::Win32::desktop_dc, LOGPIXELSY), 72);
    if (pf->italic) logfont.lfItalic = 1;
    if (pf->bold) logfont.lfWeight = FW_BOLD;
    logfont.lfCharSet = ANSI_CHARSET;

    JvGetStringUTFRegion(pf->name, 0, min(31, JvGetStringUTFLength(pf->name)), logfont.lfFaceName);
    logfont.lfFaceName[min(31, JvGetStringUTFLength(pf->name))] = 0;

    ret->hfont = (jint)CreateFontIndirect(&logfont);
    SelectObject((HDC)desktop_dc, (HFONT)(ret->hfont));

    TEXTMETRIC tm;
    GetTextMetrics((HDC)desktop_dc, &tm);
    POINT p;
    p.x = 0; p.y = tm.tmAscent;
    LPtoDP((HDC)desktop_dc, &p, 1); 
    ret->maxAscent = p.y;

    p.x = 0; p.y = tm.tmDescent;
    LPtoDP((HDC)desktop_dc, &p, 1);
    ret->maxDescent = p.y;

    return ret;
}

jint org::xwt::plat::Win32::_stringWidth(jstring font, jstring text) {

    HFONT hfont = (HFONT)(getFont(font)->hfont);
    SelectObject((HDC)org::xwt::plat::Win32::desktop_dc, hfont);

    int len = min(1024, JvGetStringUTFLength(text));
    char buf[len + 1];
    buf[len] = '\0';
    JvGetStringUTFRegion(text, 0, len, buf);
    
    SIZE size;
    GetTextExtentPoint32((HDC)org::xwt::plat::Win32::desktop_dc, buf, len, &size);
    return size.cx;
}

jboolean org::xwt::plat::Win32::_newBrowserWindow_(jstring url) {

    int len = min(2048, JvGetStringUTFLength(url));
    char buf[len + 1];
    JvGetStringUTFRegion(url, 0, len, buf);
    buf[len] = '\0';

    SHELLEXECUTEINFO ei;
    memset(&ei, 0, sizeof(ei));
    ei.cbSize = sizeof(ei);
    ei.lpVerb = "open";
    ei.lpFile = buf;
    ei.fMask  = SEE_MASK_NOCLOSEPROCESS;
    ei.nShow  = SW_SHOWDEFAULT;
    return (ShellExecuteEx(&ei) == 0);

}


// Win32DoubleBuffer /////////////////////////////////////////////////////////////////////////

// This is a scratch area; when blitting a translucent image, we copy the underlying data here first.
// Since all drawing operations are single-threaded, it's safe to use a global here.
static HBITMAP scratch = NULL;
static HDC scratch_dc = NULL;
static jint* scratch_bits = NULL;
static jint scratch_w = 0;
static jint scratch_h = 0;

#define BLT(dest, dx1, dy1, dx2, dy2, src, sx1, sy1, sx2, sy2, op)                                   \
    if ((dx2 - dx1 == sx2 - sx1) && (dy2 - dy1 == sy2 - sy1))                                        \
        BitBlt(dest, dx1, dy1, dx2 - dx1, dy2 - dy1, src, sx1, sy1, op);                             \
    else                                                                                             \
        StretchBlt(dest, dx1, dy1, dx2 - dx1, dy2 - dy1, src, sx1, sy1, sx2 - sx1, sy2 - sy1, op);
        
void org::xwt::plat::Win32$Win32DoubleBuffer::drawPicture(org::xwt::Picture* source0,
                                                          jint dx1, jint dy1, jint dx2, jint dy2,
                                                          jint sx1, jint sy1, jint sx2, jint sy2) {
    org::xwt::plat::Win32$Win32Picture* source = (org::xwt::plat::Win32$Win32Picture*)source0;

    if (source->hasalpha) {

        if (scratch == NULL || scratch_w < dx2 - dx1 || scratch_h < dy2 - dy1) {
            if (scratch_dc != NULL) DeleteDC(scratch_dc);
            if (scratch != NULL) DeleteObject(scratch);
            scratch_w = max(dx2 - dx1, scratch_w);
            scratch_h = max(dy2 - dy1, scratch_h);

            BITMAPINFO bitmapinfo;
            memset(&bitmapinfo, 0, sizeof(BITMAPINFO));
            bitmapinfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
            bitmapinfo.bmiHeader.biWidth = scratch_w;
            bitmapinfo.bmiHeader.biHeight = -1 * scratch_h;
            bitmapinfo.bmiHeader.biPlanes = 1;
            bitmapinfo.bmiHeader.biBitCount = 32;
            bitmapinfo.bmiHeader.biCompression = BI_RGB;

            // create section DIB
            scratch = CreateDIBSection(NULL, &bitmapinfo, DIB_RGB_COLORS, (void**)&scratch_bits, NULL, 0);
            scratch_dc = CreateCompatibleDC((HDC)org::xwt::plat::Win32::desktop_dc);
            SelectObject(scratch_dc, scratch);
        }

        // copy from screen to scratch
        BitBlt((HDC)scratch_dc, 0, 0, dx2 - dx1, dy2 - dy1, (HDC)hdc, dx1, dy1, SRCCOPY);

        // apply alpha-blending to scratch
        jint* dat = elements(source->data);

        for(int x = max(clipx1, dx1) - dx1; x < min(clipx2, dx2) - dx1; x++)
            for(int y = max(clipy1, dy1) - dy1; y < min(clipy2, dy2) - dy1; y++) {
                int sx = (x * (sx2 - sx1)) / (dx2 - dx1) + sx1;
                int sy = (y * (sy2 - sy1)) / (dy2 - dy1) + sy1;
                jint dst = scratch_bits[y * scratch_w + x];
                jint src = dat[sy * source->getWidth() + sx];
                jint alpha = (src & 0xFF000000) >> 24;
                jint r = (((src & 0x00FF0000) >> 16) * alpha + ((dst & 0x00FF0000) >> 16) * (0xFF - alpha)) / 0xFF;
                jint g = (((src & 0x0000FF00) >> 8)  * alpha + ((dst & 0x0000FF00) >> 8)  * (0xFF - alpha)) / 0xFF;
                jint b = (((src & 0x000000FF))       * alpha + ((dst & 0x000000FF))       * (0xFF - alpha)) / 0xFF;
                scratch_bits[y * scratch_w + x] = (r << 16) | (g << 8) | b;
            }

        // copy back from scratch to screen
        BitBlt((HDC)hdc, dx1, dy1, dx2 - dx1, dy2 - dy1, (HDC)scratch_dc, 0, 0, SRCCOPY);

    } else {
        if (source->hasmask) {
            BLT((HDC)hdc, dx1, dy1, dx2, dy2, (HDC)source->maskdc, sx1, sy1, sx2, sy2, SRCAND);
            BLT((HDC)hdc, dx1, dy1, dx2, dy2, (HDC)source->hdc, sx1, sy1, sx2, sy2, SRCPAINT);
        } else {
            BLT((HDC)hdc, dx1, dy1, dx2, dy2, (HDC)source->hdc, sx1, sy1, sx2, sy2, SRCCOPY);
        }

    }

}

void org::xwt::plat::Win32$Win32DoubleBuffer::drawString(jstring font, jstring text, jint x, jint y, jint color) {

    org::xwt::plat::Win32$Win32Font* wf = org::xwt::plat::Win32::getFont(font);
    SelectObject((HDC)hdc, (HFONT)(wf->hfont));

    // Platform API passes us the y-pos of the bottom of the text; we need the top
    y -= wf->maxAscent;

    int len = min(1024, JvGetStringUTFLength(text));
    char buf[len + 1];
    buf[len] = '\0';
    JvGetStringUTFRegion(text, 0, len, buf);

    SetTextColor((HDC)hdc, PALETTERGB((color & 0xFF0000) >> 16, (color & 0xFF00) >> 8, color & 0xFF));
    TextOut((HDC)hdc, x, y, buf, len);
}

void org::xwt::plat::Win32$Win32DoubleBuffer::fillRect(jint x, jint y, jint x2, jint y2, jint color) {
    jint w = x2 - x;
    jint h = y2 - y;

    // sadly, the ability to change the color of a brush didn't arrive until Win2k...
    HBRUSH brush = CreateSolidBrush(PALETTERGB((color & 0xFF0000) >> 16, (color & 0xFF00) >> 8, color & 0xFF));
    if (!brush) return;
    RECT rect = { x, y, x + w, y + h };
    FillRect((HDC)hdc, &rect, brush);
    DeleteObject(brush);
}

void org::xwt::plat::Win32$Win32Surface::blit(org::xwt::DoubleBuffer* s, jint sx, jint sy, jint dx, jint dy, jint dx2, jint dy2) {
    // we create the DC lazily to get around some strange race condition in WinXP
    if (hdc == 0) hdc = (jint)GetDC((HWND)hwnd);
    BitBlt((HDC)hdc, dx, dy, dx2 - dx, dy2 - dy, (HDC)(((org::xwt::plat::Win32$Win32DoubleBuffer*)s)->hdc), sx, sy, SRCCOPY);
}

void org::xwt::plat::Win32$Win32DoubleBuffer::natInit() {
    hbitmap = (jint)CreateCompatibleBitmap((HDC)org::xwt::plat::Win32::desktop_dc, w, h);
    hdc = (jint)CreateCompatibleDC((HDC)org::xwt::plat::Win32::desktop_dc);
    SetBkMode((HDC)hdc, TRANSPARENT);
    SelectObject((HDC)hdc, (HBITMAP)hbitmap);
}

void org::xwt::plat::Win32$Win32DoubleBuffer::setClip(jint x, jint y, jint x2, jint y2) {
    clipx1 = x; clipx2 = x2; clipy1 = y; clipy2 = y2;
    HRGN hrgn = CreateRectRgn(x, y, x2, y2);
    SelectClipRgn((HDC)hdc, hrgn);
    DeleteObject(hrgn);
}

void org::xwt::plat::Win32$Win32DoubleBuffer::finalize() {
    DeleteObject((void*)hdc);
    DeleteObject((void*)hbitmap);
}



// Win32Picture /////////////////////////////////////////////////////////////////////////

void org::xwt::plat::Win32$Win32Picture::natInit() {

    BITMAPINFO bitmapinfo;
    memset(&bitmapinfo, 0, sizeof(BITMAPINFO));
    bitmapinfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER); 
    bitmapinfo.bmiHeader.biWidth = w;
    bitmapinfo.bmiHeader.biHeight = -1 * h;
    bitmapinfo.bmiHeader.biPlanes = 1;
    bitmapinfo.bmiHeader.biBitCount = 32;
    bitmapinfo.bmiHeader.biCompression = BI_RGB;

    hbitmap = (jint)CreateCompatibleBitmap((HDC)org::xwt::plat::Win32::desktop_dc, w, h);
    hdc = (jint)CreateCompatibleDC((HDC)org::xwt::plat::Win32::desktop_dc);
    SelectObject((HDC)hdc, (HBITMAP)hbitmap);
    uint32_t* dat = (uint32_t*)elements(data);
    for(int i=0; i<data->length; i++) if ((dat[i] & 0xFF000000) == 0x00000000) dat[i] = 0x00000000;
    StretchDIBits((HDC)hdc, 0, 0, w, h, 0, 0, w, h, elements(data), &bitmapinfo, DIB_RGB_COLORS, SRCCOPY);

    jint _copy[min(1024, data->length)];
    jint* copy = data->length > 1024 ? (jint*)malloc(data->length * 4) : _copy;

    memcpy(copy, elements(data), data->length * 4);
    for(int i=0; i<data->length; i++)
        if ((copy[i] & 0xFF000000) == 0x00000000) {
            hasmask = 1;
            copy[i] = 0x00FFFFFF;
        } else if ((copy[i] & 0xFF000000) == 0xFF000000) {
            copy[i] = 0x00000000;
        } else {
            hasalpha = 1;
            hasmask = 0;
            if (data->length > 1024) free(copy);
            return;
        }

    if (!hasmask) {
        if (data->length > 1024) free(copy);
        return;
    }

    //    hmask = (jint)CreateBitmap(w, h, 1, 1, NULL);
    hmask = (jint)CreateCompatibleBitmap((HDC)org::xwt::plat::Win32::desktop_dc, w, h);
    maskdc = (jint)CreateCompatibleDC(NULL);
    SelectObject((HDC)maskdc, (HBITMAP)hmask);
    StretchDIBits((HDC)maskdc, 0, 0, w, h, 0, 0, w, h, copy, &bitmapinfo, DIB_RGB_COLORS, SRCCOPY);
    if (data->length > 1024) free(copy);
}



// Win32Surface /////////////////////////////////////////////////////////////////////////////

void org::xwt::plat::Win32$Win32Surface::natInit(jboolean framed) {

    // Ask the message-handling thread to create a window for us
    PostThreadMessage((DWORD)org::xwt::plat::Win32::messagePumpThread, WM_USER + 2, (WPARAM)framed, (LPARAM)this);
    hwndCreated->block();

    // turn on incremental GC now that we have a user-visible interface
    // [[this is causing segfaults; enable it later...]]
    // GC_enable_incremental();

    ShowWindow ((HWND)hwnd, SW_SHOWDEFAULT);
}

void org::xwt::plat::Win32$Win32Surface::finalize() { /* DeleteObject((void*)hwnd); */ }
void org::xwt::plat::Win32$Win32Surface::toFront() { BringWindowToTop((HWND)hwnd); }
void org::xwt::plat::Win32$Win32Surface::toBack() { SetWindowPos((HWND)hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE); }
void org::xwt::plat::Win32$Win32Surface::_dispose() { PostMessage((HWND)hwnd, WM_USER_DISPOSE, 0, 0); }
void org::xwt::plat::Win32$Win32Surface::setInvisible(jboolean h) { ShowWindow((HWND)hwnd, h ? SW_HIDE : SW_SHOWNORMAL); }
void org::xwt::plat::Win32$Win32Surface::_setMinimized(jboolean m) { ShowWindow((HWND)hwnd, m ? SW_SHOWMINIMIZED : SW_SHOWNORMAL); }
void org::xwt::plat::Win32$Win32Surface::_setMaximized(jboolean m) { ShowWindow((HWND)hwnd, m ? SW_SHOWMAXIMIZED : SW_SHOWNORMAL); }
void org::xwt::plat::Win32$Win32Surface::postCursorChange() { PostMessage((HWND)hwnd, WM_USER_SETCURSOR, 0, 0); }

void org::xwt::plat::Win32$Win32Surface::setLocation(jint x, jint y) {
    POINT point;
    RECT rect;
    point.x = 0;
    point.y = 0;
    ClientToScreen((HWND)hwnd, &point);
    GetWindowRect((HWND)hwnd, &rect);
    SetWindowPos((HWND)hwnd, NULL, x - (point.x - rect.left), y - (point.y - rect.top), 0, 0, SWP_NOZORDER | SWP_NOSIZE);
}

void org::xwt::plat::Win32$Win32Surface::setSize(jint w, jint h) {
    RECT client_rect, window_rect;
    GetClientRect((HWND)hwnd, &client_rect);
    GetWindowRect((HWND)hwnd, &window_rect);
    int width = (window_rect.right - window_rect.left) - (client_rect.right - client_rect.left) + w;
    int height = (window_rect.bottom - window_rect.top) - (client_rect.bottom - client_rect.top) + h;
    SetWindowPos((HWND)hwnd, NULL, 0, 0, width, height, SWP_NOZORDER | SWP_NOMOVE);
}

void org::xwt::plat::Win32$Win32Surface::setTitleBarText(java::lang::String* title) {
    int len = min(1024, JvGetStringUTFLength(title));
    char buf[len + 1];
    buf[len] = '\0';
    JvGetStringUTFRegion(title, 0, len, buf);
    SetWindowText((HWND)hwnd, buf);
}

void org::xwt::plat::Win32$Win32Surface::setIcon(org::xwt::Picture* p0) {

    org::xwt::plat::Win32$Win32Picture* p = (org::xwt::plat::Win32$Win32Picture*)p0;
    int icon_width = GetSystemMetrics(SM_CXSMICON);
    int icon_height = GetSystemMetrics(SM_CYSMICON);

    // we create the DC lazily to get around some strange race condition in WinXP
    if (hdc == 0) hdc = (jint)GetDC((HWND)hwnd);

    // create the bitmap
    HBITMAP bit = CreateCompatibleBitmap((HDC)hdc, icon_width, icon_height);
    HDC memdc = CreateCompatibleDC((HDC)hdc);
    SelectObject(memdc, bit);
    BLT((HDC)memdc, 0, 0, icon_width, icon_height, (HDC)(p->hdc), 0, 0, p->getWidth(), p->getHeight(), SRCCOPY);

    // create the mask
    jint* dat = elements(p->data);
    HBITMAP bit_mask = CreateBitmap(icon_width, icon_height * 2, 1, 1, NULL);
    SelectObject(memdc, bit_mask);
    for(int x=0; x<icon_width; x++)
        for(int y=0; y<icon_height; y++) {
            int source = dat[(x * p->getWidth()) / icon_width + ((y * p->getHeight()) / icon_height) * (p->getWidth())];
            if ((source & 0xFF000000) != 0x00000000) SetPixel(memdc, x, y, PALETTERGB(0, 0, 0));
            else SetPixel(memdc, x, y, PALETTERGB(255, 255, 255));
        }

    // instantiate the icon and assign it to the window class
    ICONINFO ici;
    ici.fIcon = 1;
    ici.hbmMask = bit_mask;
    ici.hbmColor = bit;
    HICON hicon = CreateIconIndirect(&ici);
    HICON oldicon = (HICON)SetClassLong((HWND)hwnd, GCL_HICONSM, (LONG)hicon);
    if (oldicon != NULL) DestroyIcon(oldicon);
    DeleteObject(memdc);
}

static jstring keyToString(WPARAM wParam);
jint org::xwt::plat::Win32$Win32Surface::WndProc(jint _hwnd, jint _iMsg, jint _wParam, jint _lParam) {

    UINT iMsg = (UINT)_iMsg;
    WPARAM wParam = (WPARAM)_wParam;
    LPARAM lParam = (LPARAM)_lParam;

    int oldmousex, oldmousey;
    MINMAXINFO* mmi;
    POINT point;
    HWND hwnd2;
    RECT rect, rect2;
    RECT client_rect, window_rect;
    jboolean newinside;
    int16_t mouse_x;
    int16_t mouse_y;
    int addwidth, addheight;

    switch(iMsg) {
    case WM_DEVMODECHANGE: break;    // FEATURE: color depth changed
    case WM_DISPLAYCHANGE: break;    // FEATURE: screen size changed
    case WM_FONTCHANGE: break;       // FEATURE: set of fonts changed
    case WM_MOUSEWHEEL: break;       // FEATURE: Mouse Wheel

    case WM_SYSKEYDOWN: if (GetKeyState(VK_MENU) >> (sizeof(SHORT) * 8 - 1) == 0) return 0;
    case WM_KEYDOWN:
        KeyPressed(keyToString(wParam));
        return 0;

    case WM_SYSKEYUP: if (GetKeyState(VK_MENU) >> (sizeof(SHORT) * 8 - 1) == 0) return 0;
    case WM_KEYUP:
        KeyReleased(keyToString(wParam));
        return 0;

    case WM_SETFOCUS: Focused(true); return 0;
    case WM_KILLFOCUS: Focused(false); return 0;
    case WM_LBUTTONDBLCLK: DoubleClick(1); return 0;
    case WM_RBUTTONDBLCLK: DoubleClick(2); return 0;
    case WM_MBUTTONDBLCLK: DoubleClick(3); return 0;
    case WM_LBUTTONDOWN: Press(1); return 0;
    case WM_RBUTTONDOWN: Press(2); return 0;
    case WM_MBUTTONDOWN: Press(3); return 0;
    case WM_CLOSE: Close(); return 0;
    case WM_ERASEBKGND: return 0;

    case WM_LBUTTONUP:
    case WM_RBUTTONUP:
    case WM_MBUTTONUP:
        Release(iMsg == WM_LBUTTONUP ? 1 : (iMsg == WM_RBUTTONUP ? 2 : 3));
        if (captured && !inside) {
            ReleaseCapture();
            SetCursor((HCURSOR)previous_cursor);
        }
        return 0;

    case WM_MOVING:
        GetClientRect((HWND)hwnd, &rect);
        PosChange(rect.left, rect.top);
        return 0;

    case WM_SIZE:
        if (wParam == SIZE_MINIMIZED) {
            if (maximized) Maximized(false);
            Minimized(true);
        } else if (wParam == SIZE_MAXIMIZED) {
            if (minimized) Minimized(false); 
            Maximized(true);
        } else {
            if (minimized) Minimized(false); 
            if (maximized) Maximized(false);
        }
        // deliberately fall through to WM_SIZING

    case WM_SIZING:
        GetClientRect((HWND)hwnd, &rect);
        SizeChange(rect.right - rect.left, rect.bottom - rect.top);
        return 0;

    case WM_MOUSEMOVE:
        if (mousex == lParam & 0xFFFF && mousey == (lParam & 0xFFFF0000) >> 16) return 0;

        GetClientRect((HWND)hwnd, &rect);
        point.x = mouse_x = lParam & 0xFFFF;
        point.y = mouse_y = (lParam & 0xFFFF0000) >> 16;
        ClientToScreen((HWND)hwnd, &point);
        hwnd2 = WindowFromPoint(point);

        newinside = hwnd2 == (HWND)hwnd && mouse_x > 0 && mouse_y > 0 && mouse_x < (rect.right - rect.left) && mouse_y < (rect.bottom - rect.top) ? 1 : 0;

        Move(mouse_x, mouse_y);

        if (newinside && !inside) {
            inside = 1;
            SetCapture((HWND)hwnd);
            captured = 1;
            previous_cursor = (jint)GetCursor();
            PostMessage((HWND)hwnd, WM_USER_SETCURSOR, 0, 0);

        } else if (!newinside && inside) {
            inside = 0;
            if (!button1 && !button2 && !button3) {
                ReleaseCapture();
                captured = 0;
                SetCursor((HCURSOR)previous_cursor);
            }
        }

        return 0;

    case WM_USER_SETCURSOR:
        // I can't find documentation of this anywhere, but in my experience, an event must be received between an invocation
        // of SetCapture() and SetCursor(). Furthermore, it seems that after calling SetCursor(), the cursor will not change until a
        // return from WndProc (ie the completion of the processing of some event).  Hence, we use WM_USER to indicate that the screen
        // cursor should be re-set to "current_cursor".
        SetCursor((HCURSOR)current_cursor);
        return 0;

    case WM_USER_DISPOSE:
        // used to signal that we should destroy ourselves
        DestroyWindow((HWND)hwnd);
        return 0;

    case WM_GETMINMAXINFO:
        GetClientRect((HWND)hwnd, &client_rect);
        GetWindowRect((HWND)hwnd, &window_rect);
        addwidth = (window_rect.right - window_rect.left) - (client_rect.right - client_rect.left);
        addheight = (window_rect.bottom - window_rect.top) - (client_rect.bottom - client_rect.top);
        mmi = (MINMAXINFO*)lParam;
        mmi->ptMinTrackSize.x = ((uint32_t)root->dmin(0)) + addwidth;
        mmi->ptMinTrackSize.y = ((uint32_t)root->dmin(1)) + addheight;
        mmi->ptMaxTrackSize.x = min(org::xwt::plat::Win32::getScreenWidth(), ((uint32_t)root->dmax(0)) + addwidth);
        mmi->ptMaxTrackSize.y = min(org::xwt::plat::Win32::getScreenHeight(), ((uint32_t)root->dmax(1)) + addheight);
        return 0;
        
    case WM_PAINT: 
        PAINTSTRUCT ps;
        BeginPaint((HWND)org::xwt::plat::Win32$Win32Surface::hwnd, &ps);
        Dirty(ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right - ps.rcPaint.left, ps.rcPaint.bottom - ps.rcPaint.top);
        EndPaint((HWND)org::xwt::plat::Win32$Win32Surface::hwnd, &ps);
        return 0;

    default: break;
    }  

    return DefWindowProc((HWND)hwnd, iMsg, wParam, lParam);
}


// Key Translator Helper /////////////////////////////////////////////////////////////////////

static char keyarr [256] = { 0 };
static jstring keyToString(WPARAM wParam) {
    char arr[8];
    keyarr[VK_CAPITAL] = GetKeyState(VK_CAPITAL);
    keyarr[VK_LSHIFT] = GetKeyState(VK_LSHIFT);
    keyarr[VK_RSHIFT] = GetKeyState(VK_RSHIFT);
    keyarr[VK_SHIFT] = GetKeyState(VK_SHIFT);

    if (ToAsciiEx(wParam, 0, (BYTE*)keyarr, (WORD*)arr, 0, GetKeyboardLayout(0)) == 1) {
        switch (arr[0]) {
        case '\t': return JvNewStringLatin1("tab");
        case 0x1b: return JvNewStringLatin1("escape");
        case '\n': return JvNewStringLatin1("enter");
        case '\r': return JvNewStringLatin1("enter");
        case 0x08: return JvNewStringLatin1("back_space");
        default: return JvNewStringLatin1(arr, 1);
        }
        
    } else switch (wParam) {
    case VK_CLEAR: return JvNewStringLatin1("clear");
    case VK_SHIFT: return JvNewStringLatin1("shift");
    case VK_CONTROL: return JvNewStringLatin1("control");
    case VK_CAPITAL: return JvNewStringLatin1("caps_lock");
    case VK_MENU: return JvNewStringLatin1("alt");
    case VK_PAUSE: return JvNewStringLatin1("pause");
    case VK_PRIOR: return JvNewStringLatin1("page_up");
    case VK_NEXT: return JvNewStringLatin1("page_down");
    case VK_END: return JvNewStringLatin1("end");
    case VK_HOME: return JvNewStringLatin1("home");
    case VK_LEFT: return JvNewStringLatin1("left");
    case VK_UP: return JvNewStringLatin1("up");
    case VK_RIGHT: return JvNewStringLatin1("right");
    case VK_DOWN: return JvNewStringLatin1("down");
    case VK_INSERT: return JvNewStringLatin1("insert");
    case VK_DELETE: return JvNewStringLatin1("delete");
    case VK_NUMPAD0: return JvNewStringLatin1("numpad0");
    case VK_NUMPAD1: return JvNewStringLatin1("numpad1");
    case VK_NUMPAD2: return JvNewStringLatin1("numpad2");
    case VK_NUMPAD3: return JvNewStringLatin1("numpad3");
    case VK_NUMPAD4: return JvNewStringLatin1("numpad4");
    case VK_NUMPAD5: return JvNewStringLatin1("numpad5");
    case VK_NUMPAD6: return JvNewStringLatin1("numpad6");
    case VK_NUMPAD7: return JvNewStringLatin1("numpad7");
    case VK_NUMPAD8: return JvNewStringLatin1("numpad8");
    case VK_NUMPAD9: return JvNewStringLatin1("numpad9");
    case VK_F1: return JvNewStringLatin1("f1");
    case VK_F2: return JvNewStringLatin1("f2");
    case VK_F3: return JvNewStringLatin1("f3");
    case VK_F4: return JvNewStringLatin1("f4");
    case VK_F5: return JvNewStringLatin1("f5");
    case VK_F6: return JvNewStringLatin1("f6");
    case VK_F7: return JvNewStringLatin1("f7");
    case VK_F8: return JvNewStringLatin1("f8");
    case VK_F9: return JvNewStringLatin1("f9");
    case VK_F10: return JvNewStringLatin1("f10");
    case VK_F11: return JvNewStringLatin1("f11");
    case VK_F12: return JvNewStringLatin1("f12");
    case VK_NUMLOCK: return JvNewStringLatin1("num_lock");
    case VK_SCROLL: return JvNewStringLatin1("scroll_lock");
    case VK_LSHIFT: return JvNewStringLatin1("shift");
    case VK_RSHIFT: return JvNewStringLatin1("shift");
    case VK_LCONTROL: return JvNewStringLatin1("control");
    case VK_RCONTROL: return JvNewStringLatin1("control");
    }
    return NULL;
}
